pub use benchmark_result::BenchmarkResult;
pub use file_to_benchmark::FileToBenchmark;

mod benchmark_result;
mod file_to_benchmark;
mod row_building;

use crate::{cli::BenchmarkCli, solving_given::solve_given, AppResult};
use anyhow::anyhow;
use colonnade::Colonnade;
use std::time::{Duration, Instant};

fn calc_average(sum: Duration, nanos_count: u64) -> Duration {
    let average_raw = sum.div_duration_f64(Duration::from_nanos(nanos_count));
    Duration::from_nanos(average_raw.round() as u64)
}
pub fn execute_benchmark(args: &BenchmarkCli) -> AppResult<String> {
    let loaded = load_benchmarks(args)?;
    let benchmarked = loaded
        .into_iter()
        .map(solve_and_keep_track_of_runtime)
        .collect::<AppResult<Vec<BenchmarkResult>>>()?;
    let sum: Duration = benchmarked.iter().map(|result| result.how_long()).sum();
    let average = calc_average(sum, benchmarked.len() as u64);
    let header = row_building::create_header();
    let after_header = row_building::create_rows_for_every_solutions(benchmarked);
    let table: Vec<Vec<String>> = after_header
        .into_iter()
        .chain(row_building::create_sum_row(sum))
        .chain(row_building::create_average_row(average))
        .chain(header)
        .collect();
    let lines = Colonnade::new(5, 100).unwrap().tabulate(table)?;

    Ok(lines.join("\n"))
}

fn solve_and_keep_track_of_runtime(
    (benchmark, content): (FileToBenchmark, String),
) -> AppResult<BenchmarkResult> {
    let (day, task) = (benchmark.given_day(), benchmark.given_task());
    let before = Instant::now();
    let actual_ouput = solve_given(day, task, &content)?;
    let after = Instant::now();

    let how_long = after - before;

    let result = BenchmarkResult::new(benchmark, actual_ouput, how_long);
    Ok(result)
}

fn load_benchmarks(args: &BenchmarkCli) -> AppResult<Vec<(FileToBenchmark, String)>> {
    let benchmarks = args
        .files()
        .iter()
        .map(|path| {
            let content = std::fs::read_to_string(path).map_err(|error| {
                anyhow!(
                    "Could not read benchmark file at {:?}\n\
                    Details: {}",
                    path,
                    error
                )
            })?;
            let parsed: Vec<FileToBenchmark> = ron::from_str(&content).map_err(|error| {
                anyhow!(
                    "Content of benchmark file at ({:?}) is in an invalid ron format\n\
                Details of invalid format: {}",
                    path,
                    error
                )
            })?;
            Ok(parsed)
        })
        .collect::<AppResult<Vec<Vec<FileToBenchmark>>>>()?
        .into_iter()
        .flatten();

    benchmarks
        .into_iter()
        .map(|benchmark| {
            let where_to_look = benchmark.where_to_look();
            let content = std::fs::read_to_string(where_to_look).map_err(|error| {
                anyhow!(
                    "Could not read puzzle input file at {:?} for day {} and task {}\n\
                    Details: {}",
                    where_to_look,
                    benchmark.given_day(),
                    benchmark.given_task(),
                    error
                )
            })?;
            Ok((benchmark, content))
        })
        .collect::<AppResult<Vec<(FileToBenchmark, String)>>>()
}
